@@ -1,7 +1,8 @@
 /*
   Copyright (C) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003,
   2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014,
-  2015 Free Software Foundation, Inc.
+  2015, 2016, 2017, 2018, 2019, 2020, 2021 Free Software Foundation,
+  Inc.
 
   This file is part of GNU Inetutils.
 
@@ -262,10 +263,19 @@ setpeer (int argc, char **argv)
 
   if (argc == 3)
     {
-      port = atoi (argv[2]);
+      if (isdigit(argv[2][0]) || argv[2][0] == '-')
+	port = atoi (argv[2]);
+      else
+	{
+	  struct servent *sp;
+
+	  sp = getservbyname (argv[2], "tcp");
+	  port = (sp) ? ntohs (sp->s_port) : 0;
+	}
+
       if (port <= 0 || port > 65535)
 	{
-	  printf ("%s: bad port number-- %s\n", argv[1], argv[2]);
+	  printf ("%s: bad port -- %s\n", argv[1], argv[2]);
 	  printf ("usage: %s host-name [port]\n", argv[0]);
 	  code = -1;
 	  return;
@@ -354,7 +364,7 @@ setpeer (int argc, char **argv)
 	    unix_server = 0;
 	  if (overbose && !strncmp (reply_string, "215 TOPS20", 10))
 	    printf
-	      ("Remember to set tenex mode when transfering binary files from this machine.\n");
+	      ("Remember to set tenex mode when transferring binary files from this machine.\n");
 	}
       verbose = overbose;
 #endif /* (unix || __unix || __unix__) && (NBBY == 8) */
@@ -1028,7 +1038,7 @@ remglob (char **argv, int doswitch)
 	args = argv;
       if ((cp = *++args) == NULL)
 	args = NULL;
-      return cp ? 0 : strdup (cp);
+      return cp ? strdup (cp) : NULL;
     }
   if (ftemp == NULL)
     {
@@ -1233,10 +1243,12 @@ sethash (int argc _GL_UNUSED_PARAMETER, char **argv)
 	{
 	case 'g':
 	case 'G':
-	  hashbytes *= 1024;	/* Cascaded multiplication!  */
+	  hashbytes *= 1024 * 1024 * 1024;
+	  break;
 	case 'm':
 	case 'M':
-	  hashbytes *= 1024;
+	  hashbytes *= 1024 * 1024;
+	  break;
 	case 'k':
 	case 'K':
 	  hashbytes *= 1024;
@@ -1712,6 +1724,9 @@ shell (int argc, char **argv _GL_UNUSED_PARAMETER)
 void
 user (int argc, char **argv)
 {
+#if !HAVE_DECL_GETPASS
+  extern char *getpass ();
+#endif
   char acct[80];
   int n, aflag = 0;
 
@@ -2061,6 +2076,9 @@ globulize (char *cp)
 void
 account (int argc, char **argv)
 {
+#if !HAVE_DECL_GETPASS
+  extern char *getpass ();
+#endif
   char acct[50], *ap;
 
   if (argc > 1)
@@ -2385,6 +2403,7 @@ domap (char *name)
 	      break;
 	    }
 	  /* Fall through, as '$' must be used verbatim.  */
+	  /* FALLTHROUGH */
 	default:
 	  if (*cp2 != *cp1)
 	    {
