@@ -1,6 +1,7 @@
 /* argcv.c - simple functions for parsing input based on whitespace
   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
-  2010, 2011, 2012 Free Software Foundation, Inc.
+  2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021
+  Free Software Foundation, Inc.
 
   This file is part of GNU Inetutils.
 
@@ -78,7 +79,7 @@ argcv_get (const char *command, const char *delim, int *argc, char ***argv)
   *argc = 0;
   *argv = NULL;
 
-  while (len > 0 && isspace (command[len - 1]))
+  while (len > 0 && isspace ((int) command[len - 1]))
     len--;
   if (len < 1)
     return 1;
