@@ -1,7 +1,7 @@
 /*
   Copyright (C) 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
-  2006, 2007, 2008, 2009, 2010, 2011, 2012 Free Software Foundation,
-  Inc.
+  2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
+  2017, 2018, 2019, 2020, 2021 Free Software Foundation, Inc.
 
   This file is part of GNU Inetutils.
 
@@ -20,7 +20,7 @@
 
 #include <config.h>
 
-#if HAVE_UTMPX_H && defined SOLARIS
+#if defined HAVE_UPDWTMPX && !defined HAVE_LOGWTMP
 # include <utmpx.h>	/* updwtmp() */
 #endif
 
@@ -30,33 +30,47 @@
 /* Solaris does not provide logwtmp(3c).
  * It is needed in addition to logwtmp_keep_open().
  */
-#if HAVE_UPDWTMP && defined SOLARIS
+#if defined HAVE_UPDWTMPX && !defined HAVE_LOGWTMP
 void
 logwtmp (char *line, char *name, char *host)
 {
-  struct utmp ut;
-#ifdef HAVE_STRUCT_UTMP_UT_TV
+  struct utmpx ut;
   struct timeval tv;
-#endif
 
   memset (&ut, 0, sizeof (ut));
-#ifdef HAVE_STRUCT_UTMP_UT_TYPE
-  ut.ut_type = USER_PROCESS;
-#endif
+#ifdef HAVE_STRUCT_UTMPX_UT_TYPE
+  if (name && *name)
+    ut.ut_type = USER_PROCESS;
+  else
+    ut.ut_type = DEAD_PROCESS;
+#endif /* UT_TYPE */
+
   strncpy (ut.ut_line, line, sizeof ut.ut_line);
+#ifdef HAVE_STRUCT_UTMPX_UT_USER
+  strncpy (ut.ut_user, name, sizeof ut.ut_user);
+#elif defined HAVE_STRUCT_UTMPX_UT_NAME
   strncpy (ut.ut_name, name, sizeof ut.ut_name);
-#ifdef HAVE_STRUCT_UTMP_UT_HOST
+#endif
+#ifdef HAVE_STRUCT_UTMPX_UT_HOST
   strncpy (ut.ut_host, host, sizeof ut.ut_host);
+# ifdef HAVE_STRUCT_UTMPX_UT_SYSLEN
+  if (strlen (host) < sizeof (ut.ut_host))
+    ut.ut_syslen = strlen (host) + 1;
+  else
+    {
+      ut.ut_host[sizeof (ut.ut_host) - 1] = '\0';
+      ut.ut_syslen = sizeof (ut.ut_host);
+    }
+# endif /* UT_SYSLEN */
+#endif
+#ifdef HAVE_STRUCT_UTMPX_UT_PID
+  ut.ut_pid = getpid ();
 #endif
 
-# ifdef HAVE_STRUCT_UTMP_UT_TV
   gettimeofday (&tv, NULL);
   ut.ut_tv.tv_sec = tv.tv_sec;
   ut.ut_tv.tv_usec = tv.tv_usec;
-# else
-  time (&ut.ut_time);
-# endif
 
-  updwtmp (PATH_WTMP, &ut);
+  updwtmpx (PATH_WTMPX, &ut);
 }
-#endif /* HAVE_UPDWTMP && defined SOLARIS */
+#endif /* HAVE_UPDWTMPX && !HAVE_LOGWTMP */
