@@ -1,6 +1,7 @@
 /*
   Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008,
-  2009, 2010, 2011, 2012 Free Software Foundation, Inc.
+  2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019,
+  2020, 2021 Free Software Foundation, Inc.
 
   This file is part of GNU Inetutils.
 
@@ -55,7 +56,7 @@
 #include <sys/stat.h>
 
 #include <errno.h>
-#include "fts.h"
+#include <fts_.h>
 #include <grp.h>
 #include <pwd.h>
 #include <stdio.h>
@@ -63,7 +64,6 @@
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
-#include <utmp.h>
 #include <filemode.h>
 
 #ifdef HAVE_SYS_MKDEV_H
@@ -133,12 +133,13 @@ printlong (DISPLAY *dp)
 		dp->s_block, (long long) howmany (sp->st_blocks, blocksize));
       strmode (sp->st_mode, buf);
       np = p->fts_pointer;
-      printf ("%s %*u %-*s  %-*s  ", buf, dp->s_nlink,
-	      sp->st_nlink, dp->s_user, np->user, dp->s_group, np->group);
+      printf ("%s %*d %-*s  %-*s  ",
+	      buf, dp->s_nlink, (int) sp->st_nlink,
+	      dp->s_user, np->user, dp->s_group, np->group);
       if (f_flags)
 	printf ("%-*s ", dp->s_flags, np->flags);
       if (S_ISCHR (sp->st_mode) || S_ISBLK (sp->st_mode))
-	printf ("%3d, %3d ", major (sp->st_rdev), minor (sp->st_rdev));
+	printf ("%3d, %3d ", (int) major (sp->st_rdev), (int) minor (sp->st_rdev));
       else if (dp->bcfile)
 	printf ("%*s%*llu ",
 		8 - dp->s_size, "", dp->s_size, (long long) sp->st_size);
